iAsc len = (num02 len ++ "/asc", [(1::Int)..2^len])
iRnd len = (num02 len ++ "/rnd", take (2^len) $ randomRs (1::Int, 100*(2^len)) $ mkStdGen 1673)

sAsc sl len = (num02 len ++ "/asc" ++ num02 sl, gen (2^len) (replicate sl 'a'))
  where gen 0 _ = []
        gen n cs = cs `seq` cs : gen (n-1) (snd $ inc cs)
        inc ['z'] = (True, ['a'])
        inc [c] = (False, [succ c])
        inc (c:cs) = case inc cs of (False, cs') -> (False, c : cs')
                                    (True, cs') -> if c == 'z' then (True, 'a' : cs')
                                                               else (False, succ c : cs')
sRnd sl len = (num02 len ++ "/rnd" ++ num02 sl, gen (2^len) (mkStdGen 1673))
  where gen num g = split num $ map chr $ randomRs (32::Int, 126) g
        split 0 _  = []
        split n cs = case splitAt sl cs of (str, cs') -> str `seq` str : split (n-1) cs'

bsAsc sl len = case sAsc sl len of (name, inputs) -> (name, map pack inputs)
bsRnd sl len = case sRnd sl len of (name, inputs) -> (name, map pack inputs)
instance NFData ByteString where rnf str = str `seq` ()

bFromList (x:xs) = foldl' bInsert bEmpty xs

num02 n | n < 10 = '0' : (show n)
        | otherwise = (show n)

---------------
-- Insert bench
---------------
benchInserts n inputs = Criterion.bgroup ("insert-" ++ n) $ map benchInsert inputs
benchInsert (name, input) = case rnf input of
  () -> Criterion.bench name $ Criterion.whnf (bForce . bFromList) input

---------------
-- Delete bench
---------------
benchDeletes n inputs = Criterion.bgroup ("delete-" ++ n) $ map benchDelete inputs
benchDelete (name, input) = case bFromList input of
  mp -> Criterion.bench name $ Criterion.whnf (bForce . bDeletes mp) input
 where bDeletes mp xs = foldl' bDelete mp xs

---------------
-- Lookup bench
---------------
countBool list = cnt list (0::Int)
  where cnt []         acc  = acc
        cnt (False:xs) acc  = cnt xs acc
        cnt (True:xs)  acc  = cnt xs (acc +1)

benchLookups n inputs = Criterion.bgroup ("lookup-" ++ n) $ map benchLookup inputs
benchLookup (name, input) = case bFromList input of
  m -> Criterion.bench name $ Criterion.whnf (countBool . map (bMember m)) input

------------------
-- Big union bench
------------------
halve [] = ([], [])
halve (x:xs) = case halve xs of (ys, zs) -> (x:zs, ys)

benchBigUnions n inputs = Criterion.bgroup ("union-" ++ n) $ map benchBigUnion inputs
benchBigUnion (name, input) = case halve input of
  (xs, ys) -> case map bFromList [xs, ys] of
    [m, n] -> Criterion.bench name $ Criterion.whnf (bForce . (uncurry bUnion)) (m, n)


-------------
-- All benchs
-------------
main = do
  args <- getArgs
  let benchname = args !! (fromJust (elemIndex "-p" args) + 1)
      lens = case (drop (length benchname - 3) benchname) of
               "int" -> [4, 12, 20]
               "str" -> [4, 10, 16]
               "bst" -> [4, 10, 16]
  Progression.defaultMain $ Criterion.bgroup "" $ []
#if !defined(BENCH_STR_ONLY) && !defined(BENCH_BSTR_ONLY)
    ++ case benchname of
      "insert-int" -> [benchInserts "int" [fun len | len <- lens, fun <- [ iAsc, iRnd ]]]
      "delete-int" -> [benchDeletes "int" [fun len | len <- lens, fun <- [ iAsc, iRnd ]]]
      "lookup-int" -> [benchLookups "int" [fun len | len <- lens, fun <- [ iAsc, iRnd ]]]
      "union-int"  -> [benchBigUnions "int" [fun len | len<-lens, fun <- [ iAsc, iRnd ]]]
      _            -> []
#endif
#if !defined(BENCH_INT_ONLY) && !defined(BENCH_BSTR_ONLY)
    ++ case benchname of
      "insert-str" -> [benchInserts "str" [fun len | len <- lens, fun <- [ sAsc 5,sAsc 20,sAsc 50,sRnd 5,sRnd 50]]]
      "delete-str" -> [benchDeletes "str" [fun len | len <- lens, fun <- [ sAsc 5,sAsc 20,sAsc 50,sRnd 5,sRnd 50]]]
      "lookup-str" -> [benchLookups "str" [fun len | len <- lens, fun <- [ sAsc 5,sAsc 20,sAsc 50,sRnd 5,sRnd 50]]]
      "union-str"  -> [benchBigUnions "str" [fun len | len<-lens, fun <- [ sAsc 5,sAsc 20,sAsc 50,sRnd 5,sRnd 50]]]
      _            -> []
#endif
#if !defined(BENCH_INT_ONLY) && !defined(BENCH_STR_ONLY)
    ++ case benchname of
      "insert-bst" -> [benchInserts "bst" [fun len | len <- lens, fun <- [bsAsc 5,bsAsc 20,bsAsc 50,bsRnd 5,bsRnd 50]]]
      "delete-bst" -> [benchDeletes "bst" [fun len | len <- lens, fun <- [bsAsc 5,bsAsc 20,bsAsc 50,bsRnd 5,bsRnd 50]]]
      "lookup-bst" -> [benchLookups "bst" [fun len | len <- lens, fun <- [bsAsc 5,bsAsc 20,bsAsc 50,bsRnd 5,bsRnd 50]]]
      "union-bst"  -> [benchBigUnions "bst" [fun len | len<-lens, fun <- [bsAsc 5,bsAsc 20,bsAsc 50,bsRnd 5,bsRnd 50]]]
      _            -> []
#endif
