module Data.RBCSSet( RBCSet
                   , singleton
                   , insert
                   , member
                   , union
                   , toList
                   , toAscList
                   , fromList
                   , fromAscList
                   ) where

import Prelude hiding (lookup)
import Control.DeepSeq
import Control.Monad (guard)
import Data.List (sort, groupBy, foldl')
import Debug.Trace (trace)
import Test.QuickCheck hiding (quickCheck)

data RBCSet k = E
               | B {-# UNPACK #-}!Int k (RBCSet k) (RBCSet k)
               | R {-# UNPACK #-}!Int k (RBCSet k) (RBCSet k)
               deriving Show

instance NFData k => NFData (RBCSet k) where
  rnf E = ()
  rnf (B _ k l r) = rnf k `seq` rnf l `seq` rnf r
  rnf (R _ k l r) = rnf k `seq` rnf l `seq` rnf r

empty = E
singleton k = B 1 k E E

{-# NOINLINE insert #-}
insert k E = B 1 k E E
insert k tree = case ins k tree of
                    (R s k l r) -> (B s k l r)
                    tree        -> tree
  where
    ins k' E = (R 1 k' E E)
    ins k' (B s k l r) = case compare k' k of
      EQ -> B s k l r
      LT -> case l of
              (R ls lk ll lr) -> case compare k' lk of
                 EQ -> (B s k (R ls lk ll lr) r)
                 LT -> case ins k' ll of
                         nll@(R _ _ _ _) -> case r of
                                              (R rs rk rl rr) -> (doR k (doB lk nll lr) (B rs rk rl rr))
                                              r               -> (doB lk nll (doR k lr r))
                         nll             -> (doB k (doR lk nll lr) r)
                 GT -> case ins k' lr of
                         nlr@(R lrs lrk lrl lrr) -> case r of
                                                      (R rs rk rl rr) -> (doR k (doB lk ll nlr) (B rs rk rl rr))
                                                      r               -> (doB lrk (doR lk ll lrl) (doR k lrr r))
                         nlr                     -> (doB k (doR lk ll nlr) r)
              l -> (doB k (ins k' l) r)
      GT -> case r of
              (R rs rk rl rr) -> case compare k' rk of
                 EQ -> (B s k l (R rs rk rl rr))
                 GT -> case ins k' rr of
                         nrr@(R _ _ _ _) -> case l of
                                              (R ls lk ll lr) -> (doR k (B ls lk ll lr) (doB rk rl nrr))
                                              l               -> (doB rk (doR k l rl) nrr)
                         nrr             -> (doB k l (doR rk rl nrr))
                 LT -> case ins k' rl of
                         nrl@(R rls rlk rll rlr) -> case l of
                                                      (R ls lk ll lr) -> (doR k (B ls lk ll lr) (doB rk nrl rr))
                                                      l               -> (doB rlk (doR k l rll) (doR rk rlr rr))
                         nrl                     -> (doB k l (doR rk nrl rr))
              r -> (doB k l (ins k' r))

{-# INLINE size #-}
size E = 0
size (R s _ _ _) = s
size (B s _ _ _) = s

{-# INLINE doR #-}
doR k  E E = R 1 k E E
doR k  E r@(R s _ _ _) = R (s+1) k E r
doR k  E r@(B s _ _ _) = R (s+1) k E r
doR k  l@(R s _ _ _) E = R (s+1) k l E
doR k  l@(B s _ _ _) E = R (s+1) k l E
doR k  l@(R ls _ _ _) r@(R rs _ _ _) = R (ls+rs+1) k l r
doR k  l@(R ls _ _ _) r@(B rs _ _ _) = R (ls+rs+1) k l r
doR k  l@(B ls _ _ _) r@(R rs _ _ _) = R (ls+rs+1) k l r
doR k  l@(B ls _ _ _) r@(B rs _ _ _) = R (ls+rs+1) k l r

{-# INLINE doB #-}
doB k E E = B 1 k E E
doB k E r@(R s _ _ _) = B (s+1) k E r
doB k E r@(B s _ _ _) = B (s+1) k E r
doB k l@(R s _ _ _) E = B (s+1) k l E
doB k l@(B s _ _ _) E = B (s+1) k l E
doB k l@(R ls _ _ _) r@(R rs _ _ _) = B (ls+rs+1) k l r
doB k l@(R ls _ _ _) r@(B rs _ _ _) = B (ls+rs+1) k l r
doB k l@(B ls _ _ _) r@(R rs _ _ _) = B (ls+rs+1) k l r
doB k l@(B ls _ _ _) r@(B rs _ _ _) = B (ls+rs+1) k l r

{-# INLINE member #-}
member k' E = False
member k' tree = lkp tree
 where lkp E = False
       lkp (B _ k l r) = case compare k' k of
                            EQ -> True
                            LT -> lkp l
                            GT -> lkp r
       lkp (R _ k l r) = case compare k' k of
                             EQ -> True
                             LT -> lkp l
                             GT -> lkp r

{-# NOINLINE union #-}
union E x = x
union x E = x
-- TODO: Tune the constants of algorithms crossover.
union x y | size x <= 1      = case toList x of k : _ ->  insert k y
          | size y <= 1      = case toList y of k : _ ->  insert k x
          | size x <= 2*size y && size y <= 2*size x = case mergeToList x y of (list, len) -> fromAscListLen list len
          | size x <= size y = insertList (toAscList x) y
          | otherwise        = insertList (toAscList y) x

mergeToList x y = mrg (toDscList x) (toDscList y) [] (0::Int)
 where mrg _                   _              acc len | acc `seq` len `seq` False = undefined
       mrg []                  []             acc len = (acc, len)
       mrg []                  (x:xs)         acc len = mrg [] xs (x:acc) (len+1)
       mrg (x:xs)              []             acc len = mrg [] xs (x:acc) (len+1)
       mrg xS@(x : xs) yS@(y : ys) acc len = case compare x y of
                                               LT -> mrg xS ys (y:acc) (len+1)
                                               EQ -> mrg xs ys (x:acc) (len+1)
                                               GT -> mrg xs yS (x:acc) (len+1)


insertList xs tree = foldr insert tree xs

-- insertList [] tree = tree
-- insertList xs tree = case insL xs tree Nothing of
--                       ((R s k l r), xs)  -> insertList xs (B s k l r)
--                       (tree,          [])  -> tree
--                       otherwise            -> error "Error in insertList implementation"
--   where
--     insL []          tree        _       = (tree, [])
--     insL x@(k' : _ ) tree       (Just b) | k' >= b = (tree, x)
--     insL x@(k' : xs) E           b       = ((R 1 k' E E), xs)
--     insL x@(k' : xs) (B s k l r) b       = case compare k' k of
--       EQ -> insL xs (B s k l r) b
--       LT -> case l of
--               (R ls lk ll lr) -> case compare k' lk of
--                  EQ -> insL xs (B s k (R ls lk ll lr) r) b
--                  LT -> case insL x ll (Just lk) of
--                          (nll@(R _ _ _ _), xs') -> case r of
--                                                      (R rs rk rl rr) -> (doR k (doB lk nll lr) (B rs rk rl rr), xs')
--                                                      r               -> insL xs' (doB lk nll (doR k lr r)) b
--                          (nll,             xs') -> insL xs' (doB k (doR lk nll lr) r) b
--                  GT -> case insL x lr (Just k) of
--                          (nlr@(R lrs lrk lrl lrr), xs') -> case r of
--                                                              (R rs rk rl rr) -> ((doR k (doB lk ll nlr) (B rs rk rl rr)), xs')
--                                                              r               -> insL xs' (doB lrk (doR lk ll lrl) (doR k lrr r)) b
--                          (nlr,                     xs') -> insL xs' (doB k (doR lk ll nlr) r) b
--               l -> case insL x l (Just k) of (l', xs') -> insL xs' (doB k l' r) b
--       GT -> case r of
--               (R rs rk rl rr) -> case compare k' rk of
--                  EQ -> insL xs (B s k l (R rs rk rl rr)) b
--                  GT -> case insL x rr b of
--                          (nrr@(R _ _ _ _), xs') -> case l of
--                                                        (R ls lk ll lr) -> (doR k (B ls lk ll lr) (doB rk rl nrr), xs')
--                                                        l               -> insL xs' (doB rk (doR k l rl) nrr) b
--                          (nrr,               xs') -> (doB k l (doR rk rl nrr), xs')
--                  LT -> case insL x rl (Just rk) of
--                          (nrl@(R rls rlk rll rlr), xs') -> case l of
--                                                               (R ls lk ll lr) -> (doR k (B ls lk ll lr) (doB rk nrl rr), xs')
--                                                               l               -> insL xs' (doB rlk (doR k l rll) (doR rk rlr rr)) b
--                          (nrl,                      xs') -> insL xs' (doB k l (doR rk nrl rr)) b
--               r -> case insL x r b of (r', xs') -> insL xs (doB k l r') b


toList tree = toList' tree []
  where toList' E acc = acc
        toList' (B _ k l r) acc = k : toList' l (toList' r acc)
        toList' (R _ k l r) acc = k : toList' l (toList' r acc)

toAscList tree = toAscList' tree []
  where toAscList' E acc = acc
        toAscList' (B _ k l r) acc = toAscList' l (k : (toAscList' r acc))
        toAscList' (R _ k l r) acc = toAscList' l (k : (toAscList' r acc))

-- TODO: Try inlining to increase mrg speed?
toDscList tree = toDscList' tree []
  where toDscList' E acc = acc
        toDscList' (B _ k l r) acc = toDscList' r (k : (toDscList' l acc))
        toDscList' (R _ k l r) acc = toDscList' r (k : (toDscList' l acc))

fromList list = foldl' (flip insert) E list

fromAscList list = fromAscListLen list (length list)

fromAscListLen :: [k] -> Int -> RBCSet k
fromAscListLen _ 0 = E
fromAscListLen list len = fst $ listToTree len (lg (len+1)) list
  where lg 1 = 0
        lg n = 1 + lg (n `div` 2)

        listToTree 0 d xs       = d `seq` (E, xs)
        listToTree 1 1 (k : xs) = (B 1 k E E, xs)
        listToTree 1 0 (k : xs) = (R 1 k E E, xs)
        listToTree n d xs = case listToTree (n `div` 2) (d - 1) xs of
                              (l, k : xs') -> case listToTree (n - 1 - n `div` 2) (d - 1) xs' of
                                (r, xs'') -> (B n k l r, xs'')

isRBTree tree = maybe False (const True) $ isRB tree 0 Nothing Nothing
 where isRB E _ _ _ = return (1, 0)
       isRB (R s k l r) _ lb rb = do
         guard (case l of { R _ _ _ _ -> False; otherwise -> True })
         guard (case r of { R _ _ _ _ -> False; otherwise -> True })
         isRB (B s k l r) 1 lb rb
       isRB (B s k l r) c lb rb = do
         guard (maybe True (k >) lb)
         guard (maybe True (k <) rb)
         (ld, ls) <- isRB l 0 lb (Just k)
         (rd, rs) <- isRB r 0 (Just k) rb
         guard (ld == rd)
         guard (ls + rs + 1 == s)
         return (ld + 1 - c, s)

uniq list = let len = length list
            in len == (length $ groupBy (\x y->fst x == fst y) (sort list))

prop_FromList list = uniq list ==> let tree = fromList list
                                   in isRBTree tree && (sort list == toAscList tree)
 where types = list::[(Int, Int)]

prop_FromAscList list = uniq list ==> let tree = fromAscList (sort list)
                                      in isRBTree tree && (sort list == toAscList tree)
 where types = list::[(Int, Int)]

prop_Union xs ys = uniq (xs ++ ys) ==> let u = union (fromList xs) (fromList ys)
                                       in isRBTree u && (sort (xs ++ ys) == toAscList u)
 where types = (xs::[(Int, Int)], ys::[(Int, Int)])

prop_ToAscList list = uniq list ==> (sort list == (toAscList $ fromAscList $ sort list))
 where types = list::[(Int, Int)]

prop_ToDscList list = uniq list ==> (sort list == (reverse $ toDscList $ fromAscList $ sort list))
 where types = list::[(Int, Int)]

thorough = Config
  { configMaxTest = 1000
  , configMaxFail = 20000
  , configSize    = (+ 3) . (`div` 2)
  , configEvery   = \n args -> let s = show n in s ++ [ '\b' | _ <- s ]
  }

quickCheck :: (Testable a) => a -> IO ()
quickCheck = check thorough
