module Data.RBCSMap( RBCMap
                   , singleton
                   , insert
                   , lookup
                   , member
                   , union
                   , toList
                   , toAscList
                   , fromList
                   , fromAscList
                   ) where

import Prelude hiding (lookup)
import Control.DeepSeq
import Control.Monad (guard)
import Data.List (sort, groupBy)
import Debug.Trace (trace)
import Test.QuickCheck hiding (quickCheck)

data RBCMap k v = E
               | B {-# UNPACK #-}!Int k v (RBCMap k v) (RBCMap k v)
               | R {-# UNPACK #-}!Int k v (RBCMap k v) (RBCMap k v)
               deriving Show

instance (NFData k, NFData a) => NFData (RBCMap k a) where
  rnf E = ()
  rnf (B _ k v l r) = rnf k `seq` rnf v `seq` rnf l `seq` rnf r
  rnf (R _ k v l r) = rnf k `seq` rnf v `seq` rnf l `seq` rnf r

empty = E
singleton k v = B 1 k v E E

{-# NOINLINE insert #-}
insert k v E = B 1 k v E E
insert k v tree = case ins k v tree of
                    (R s k v l r) -> (B s k v l r)
                    tree          -> tree
  where
    ins k' v' E = (R 1 k' v' E E)
    ins k' v' (B s k v l r) = case compare k' k of
      EQ -> B s k v' l r
      LT -> case l of
              (R ls lk lv ll lr) -> case compare k' lk of
                 EQ -> (B s k v (R ls lk v' ll lr) r)
                 LT -> case ins k' v' ll of
                         nll@(R _ _ _ _ _) -> case r of
                                                (R rs rk rv rl rr) -> (doR k v (doB lk lv nll lr) (B rs rk rv rl rr))
                                                r                  -> (doB lk lv nll (doR k v lr r))
                         nll               -> (doB k v (doR lk lv nll lr) r)
                 GT -> case ins k' v' lr of
                         nlr@(R lrs lrk lrv lrl lrr) -> case r of
                                                          (R rs rk rv rl rr) -> (doR k v (doB lk lv ll nlr) (B rs rk rv rl rr))
                                                          r                  -> (doB lrk lrv (doR lk lv ll lrl) (doR k v lrr r))
                         nlr                         -> (doB k v (doR lk lv ll nlr) r)
              l -> (doB k v (ins k' v' l) r)
      GT -> case r of
              (R rs rk rv rl rr) -> case compare k' rk of
                 EQ -> (B s k v l (R rs rk v' rl rr))
                 GT -> case ins k' v' rr of
                         nrr@(R _ _ _ _ _) -> case l of
                                                (R ls lk lv ll lr) -> (doR k v (B ls lk lv ll lr) (doB rk rv rl nrr))
                                                l                  -> (doB rk rv (doR k v l rl) nrr)
                         nrr               -> (doB k v l (doR rk rv rl nrr))
                 LT -> case ins k' v' rl of
                         nrl@(R rls rlk rlv rll rlr) -> case l of
                                                          (R ls lk lv ll lr) -> (doR k v (B ls lk lv ll lr) (doB rk rv nrl rr))
                                                          l                  -> (doB rlk rlv (doR k v l rll) (doR rk rv rlr rr))
                         nrl                         -> (doB k v l (doR rk rv nrl rr))
              r -> (doB k v l (ins k' v' r))

{-# INLINE size #-}
size E = 0
size (R s _ _ _ _) = s
size (B s _ _ _ _) = s

{-# INLINE doR #-}
doR k v E E = R 1 k v E E
doR k v E r@(R s _ _ _ _) = R (s+1) k v E r
doR k v E r@(B s _ _ _ _) = R (s+1) k v E r
doR k v l@(R s _ _ _ _) E = R (s+1) k v l E
doR k v l@(B s _ _ _ _) E = R (s+1) k v l E
doR k v l@(R ls _ _ _ _) r@(R rs _ _ _ _) = R (ls+rs+1) k v l r
doR k v l@(R ls _ _ _ _) r@(B rs _ _ _ _) = R (ls+rs+1) k v l r
doR k v l@(B ls _ _ _ _) r@(R rs _ _ _ _) = R (ls+rs+1) k v l r
doR k v l@(B ls _ _ _ _) r@(B rs _ _ _ _) = R (ls+rs+1) k v l r

{-# INLINE doB #-}
doB k v E E = B 1 k v E E
doB k v E r@(R s _ _ _ _) = B (s+1) k v E r
doB k v E r@(B s _ _ _ _) = B (s+1) k v E r
doB k v l@(R s _ _ _ _) E = B (s+1) k v l E
doB k v l@(B s _ _ _ _) E = B (s+1) k v l E
doB k v l@(R ls _ _ _ _) r@(R rs _ _ _ _) = B (ls+rs+1) k v l r
doB k v l@(R ls _ _ _ _) r@(B rs _ _ _ _) = B (ls+rs+1) k v l r
doB k v l@(B ls _ _ _ _) r@(R rs _ _ _ _) = B (ls+rs+1) k v l r
doB k v l@(B ls _ _ _ _) r@(B rs _ _ _ _) = B (ls+rs+1) k v l r

{-# INLINE lookup #-}
lookup k' E = Nothing
lookup k' tree = lkp tree
 where lkp E = Nothing
       lkp (B _ k v l r) = case compare k' k of
                            EQ -> Just v
                            LT -> lkp l
                            GT -> lkp r
       lkp (R _ k v l r) = case compare k' k of
                             EQ -> Just v
                             LT -> lkp l
                             GT -> lkp r

{-# INLINE member #-}
member k' E = False
member k' tree = lkp tree
 where lkp E = False
       lkp (B _ k v l r) = case compare k' k of
                            EQ -> True
                            LT -> lkp l
                            GT -> lkp r
       lkp (R _ k v l r) = case compare k' k of
                             EQ -> True
                             LT -> lkp l
                             GT -> lkp r

{-# NOINLINE union #-}
union E x = x
union x E = x
-- TODO: Tune the constants of algorithms crossover.
union x y | size x <= 1      = case toList x of (k, v) : _ ->  insert k v y
          | size y <= 1      = case toList y of (k, v) : _ ->  insert k v x
          | size x <= 2*size y && size y <= 2*size x = case mergeToList x y of (list, len) -> fromAscListLen list len
          | size x <= size y = insertList (toAscList x) y
          | otherwise        = insertList (toAscList y) x

mergeToList x y = mrg (toDscList x) (toDscList y) [] (0::Int)
 where mrg _                   _              acc len | acc `seq` len `seq` False = undefined
       mrg []                  []             acc len = (acc, len)
       mrg []                  (x:xs)         acc len = mrg [] xs (x:acc) (len+1)
       mrg (x:xs)              []             acc len = mrg [] xs (x:acc) (len+1)
       mrg xS@(x@(kx,vx) : xs) yS@(y@(ky,vy) : ys) acc len = case compare kx ky of
                                                               LT -> mrg xS ys (y:acc) (len+1)
                                                               EQ -> mrg xS yS (x:acc) (len+1)
                                                               GT -> mrg xs yS (x:acc) (len+1)


insertList [] tree = tree
insertList xs tree = case insL xs tree Nothing of
                      ((R s k v l r), xs)  -> insertList xs (B s k v l r)
                      (tree,          [])  -> tree
                      otherwise            -> error "Error in insertList implementation"
  where
    insL []               tree          _            = (tree, [])
    insL x@((k',_ ) : _ ) tree    (Just b) | k' >= b = (tree, x)
    insL x@((k',v') : xs) E             b            = ((R 1 k' v' E E), xs)
    insL x@((k',v') : xs) (B s k v l r) b            = case compare k' k of
      EQ -> insL xs (B s k v' l r) b
      LT -> case l of
              (R ls lk lv ll lr) -> case compare k' lk of
                 EQ -> insL xs (B s k v (R ls lk v' ll lr) r) b
                 LT -> case insL x ll (Just lk) of
                         (nll@(R _ _ _ _ _), xs') -> case r of
                                                       (R rs rk rv rl rr) -> (doR k v (doB lk lv nll lr) (B rs rk rv rl rr), xs')
                                                       r                  -> insL xs' (doB lk lv nll (doR k v lr r)) b
                         (nll,               xs') -> insL xs' (doB k v (doR lk lv nll lr) r) b
                 GT -> case insL x lr (Just k) of
                         (nlr@(R lrs lrk lrv lrl lrr), xs') -> case r of
                                                                 (R rs rk rv rl rr) -> ((doR k v (doB lk lv ll nlr) (B rs rk rv rl rr)), xs')
                                                                 r                  -> insL xs' (doB lrk lrv (doR lk lv ll lrl) (doR k v lrr r)) b
                         (nlr,                         xs') -> insL xs' (doB k v (doR lk lv ll nlr) r) b
              l -> case insL x l (Just k) of (l', xs') -> insL xs' (doB k v l' r) b
      GT -> case r of
              (R rs rk rv rl rr) -> case compare k' rk of
                 EQ -> insL xs (B s k v l (R rs rk v' rl rr)) b
                 GT -> case insL x rr b of
                         (nrr@(R _ _ _ _ _), xs') -> case l of
                                                       (R ls lk lv ll lr) -> (doR k v (B ls lk lv ll lr) (doB rk rv rl nrr), xs')
                                                       l                  -> insL xs' (doB rk rv (doR k v l rl) nrr) b
                         (nrr,               xs') -> (doB k v l (doR rk rv rl nrr), xs')
                 LT -> case insL x rl (Just rk) of
                         (nrl@(R rls rlk rlv rll rlr), xs') -> case l of
                                                                 (R ls lk lv ll lr) -> (doR k v (B ls lk lv ll lr) (doB rk rv nrl rr), xs')
                                                                 l                  -> insL xs' (doB rlk rlv (doR k v l rll) (doR rk rv rlr rr)) b
                         (nrl,                         xs') -> insL xs' (doB k v l (doR rk rv nrl rr)) b
              r -> case insL x r b of (r', xs') -> insL xs (doB k v l r') b


toList tree = toList' tree []
  where toList' E acc = acc
        toList' (B _ k v l r) acc = (k, v) : toList' l (toList' r acc)
        toList' (R _ k v l r) acc = (k, v) : toList' l (toList' r acc)

toAscList tree = toAscList' tree []
  where toAscList' E acc = acc
        toAscList' (B _ k v l r) acc = toAscList' l ((k, v) : (toAscList' r acc))
        toAscList' (R _ k v l r) acc = toAscList' l ((k, v) : (toAscList' r acc))

-- TODO: Try inlining to increase mrg speed?
toDscList tree = toDscList' tree []
  where toDscList' E acc = acc
        toDscList' (B _ k v l r) acc = toDscList' r ((k, v) : (toDscList' l acc))
        toDscList' (R _ k v l r) acc = toDscList' r ((k, v) : (toDscList' l acc))

fromList list = foldl (\tree (k, v) -> insert k v tree) E list

fromAscList list = fromAscListLen list (length list)

fromAscListLen :: [(k, v)] -> Int -> RBCMap k v
fromAscListLen _ 0 = E
fromAscListLen list len = fst $ listToTree len (lg (len+1)) list
  where lg 1 = 0
        lg n = 1 + lg (n `div` 2)

        listToTree 0 d xs           = d `seq` (E, xs)
        listToTree 1 1 ((k,v) : xs) = (B 1 k v E E, xs)
        listToTree 1 0 ((k,v) : xs) = (R 1 k v E E, xs)
        listToTree n d xs = case listToTree (n `div` 2) (d - 1) xs of
                              (l, (k,v) : xs') -> case listToTree (n - 1 - n `div` 2) (d - 1) xs' of
                                (r, xs'') -> (B n k v l r, xs'')

isRBTree tree = maybe False (const True) $ isRB tree 0 Nothing Nothing
 where isRB E _ _ _ = return (1, 0)
       isRB (R s k v l r) _ lb rb = do
         guard (case l of { R _ _ _ _ _ -> False; otherwise -> True })
         guard (case r of { R _ _ _ _ _ -> False; otherwise -> True })
         isRB (B s k v l r) 1 lb rb
       isRB (B s k v l r) c lb rb = do
         guard (maybe True (k >) lb)
         guard (maybe True (k <) rb)
         (ld, ls) <- isRB l 0 lb (Just k)
         (rd, rs) <- isRB r 0 (Just k) rb
         guard (ld == rd)
         guard (ls + rs + 1 == s)
         return (ld + 1 - c, s)

uniq list = let len = length list
            in len == (length $ groupBy (\x y->fst x == fst y) (sort list))

prop_FromList list = uniq list ==> let tree = fromList list
                                   in isRBTree tree && (sort list == toAscList tree)
 where types = list::[(Int, Int)]

prop_FromAscList list = uniq list ==> let tree = fromAscList (sort list)
                                      in isRBTree tree && (sort list == toAscList tree)
 where types = list::[(Int, Int)]

prop_Union xs ys = uniq (xs ++ ys) ==> let u = union (fromList xs) (fromList ys)
                                       in isRBTree u && (sort (xs ++ ys) == toAscList u)
 where types = (xs::[(Int, Int)], ys::[(Int, Int)])

prop_ToAscList list = uniq list ==> (sort list == (toAscList $ fromAscList $ sort list))
 where types = list::[(Int, Int)]

prop_ToDscList list = uniq list ==> (sort list == (reverse $ toDscList $ fromAscList $ sort list))
 where types = list::[(Int, Int)]

thorough = Config
  { configMaxTest = 1000
  , configMaxFail = 20000
  , configSize    = (+ 3) . (`div` 2)
  , configEvery   = \n args -> let s = show n in s ++ [ '\b' | _ <- s ]
  }

quickCheck :: (Testable a) => a -> IO ()
quickCheck = check thorough
